from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class SerializerProtocol(Protocol):  # pragma: no cover
    def dump(self, *args: Any, **kwargs: Any) -> None: ...

    def dumps(self, *args: Any, **kwargs: Any) -> Any: ...

    def load(self, *args: Any, **kwargs: Any) -> Any: ...

    def loads(self, *args: Any, **kwargs: Any) -> Any: ...
